-- Prompt lazy assertions based on monadic parser combinators
-- As described in APLAS 2007
-- Version 20/11/2007

module MonadicAssertions 
  (Try, Assert, Lazy, assert
  ,notAssert, (<|>),(<$>),(&&&),(<&>),(&>),(==>)
  ,pVal,pPrim,p0,p1,p2,p3,pFun
  ,fun1,fun2,fun3,pNil,pCons,pTuple
  ,allA
  ,check1,check2
  ,Observe(obs),oPrim,o0,o1,o2,o3) 
  where


import Observer
import PosTrie
import Monad (guard)
import Try

infixl 3  <&>, &>, &&&


-- additional monadic operators:

-- Succeed iff both parsers succeed independently 
-- but don't care about their results.
(&&&) :: Try a -> Try b -> Try ()
t1 &&& t2 = return (const (const ())) <$> t1 <$> t2

-- Succeed iff both parsers succeed independently
-- and return both results
(<&>) :: Try a -> Try b -> Try (a,b)
xs <&> ys = return (,) <$> xs <$> ys

-- Parse both but return only result of second parser.
(&>) :: Try a -> Try b -> Try b
x &> y = return (const id) <$> x <$> y

-- Implication: if first parser succeeds then second one needs to succeed.
(==>) :: Try a -> Try () -> Try ()
x ==> y = notAssert x <|> y

-- Apply non-monadic function to any lazy value.
check1 :: (a -> Bool) -> Lazy a -> Try ()
check1 p lx = do 
  x <- pVal lx
  guard (p x)

-- Apply non-monadic function to any two lazy values.
check2 :: (a -> b -> Bool) -> Lazy a -> Lazy b -> Try ()
check2 p lx ly = do
  x <- pVal lx
  y <- pVal ly
  guard (p x y)

-- -------------------------------------
-- parsers for specific types

pNil :: Lazy [a] -> Try ()
pNil = p0 null

pCons :: Lazy [a] -> Try (Lazy a, Lazy [a])
pCons = p2 (\xs -> case xs of
                     (y:ys) -> Just (y,ys)
                     []     -> Nothing)

pTuple :: Lazy (a,b) -> Try (Lazy a,Lazy b)
pTuple = p2 Just

fun3 :: (Lazy a -> Lazy b -> Lazy c -> Lazy d -> Assert) ->
        Lazy (a -> b -> c -> d) -> Assert
fun3 = liftFun fun2

fun2 :: (Lazy a -> Lazy b -> Lazy c -> Assert) -> Lazy (a -> b -> c) -> Assert
fun2 = liftFun fun1

fun1 :: (Lazy a -> Lazy b -> Assert) -> Lazy (a -> b) -> Assert
fun1 = liftFun id 

liftFun :: (b -> Lazy c -> Assert) -> (Lazy a -> b) -> Lazy (a -> c) -> Assert
liftFun g f lh = pFun lh >>= allA (\xy -> pTuple xy >>= \(x,y) -> g (f x) y)

allA :: (Lazy a -> Assert) -> Lazy [a] -> Assert
allA p xs =
    pNil xs -- seems to be not so intuitive for fun lists; special version for
            -- infinite list?
  <|> do
    (y,ys) <- pCons xs
    (p y &> allA p ys)

-- -------------------------------------
-- end


